(* chapter 'Biegelinie' from the textbook: 
   Timischl, Kaiser. Ingenieur-Mathematik 3. Wien 1999. p.268-271.
   author: Walther Neuper
   050826,
   (c) due to copyright terms

remove_thy"Biegelinie";
use_thy"Knowledge/Biegelinie";
use_thy_only"Knowledge/Biegelinie";

remove_thy"Biegelinie";
use_thy"Knowledge/Isac";
*)

Biegelinie = Integrate + Equation + EqSystem +

consts

  q_    :: real => real ("q'_")     (* Streckenlast               *)
  Q     :: real => real             (* Querkraft                  *)
  Q'    :: real => real             (* Ableitung der Querkraft    *)
  M'_b  :: real => real ("M'_b")    (* Biegemoment                *)
  M'_b' :: real => real ("M'_b'")   (* Ableitung des Biegemoments *)
  y''   :: real => real             (* 2.Ableitung der Biegeline  *)
  y'    :: real => real             (* Neigung der Biegeline      *)
(*y     :: real => real             (* Biegeline                  *)*)
  EI    :: real                     (* Biegesteifigkeit           *)

  (*new Descriptions in the related problems*)
  Traegerlaenge            :: real => una
  Streckenlast             :: real => una
  BiegemomentVerlauf       :: bool => una
  Biegelinie               :: (real => real) => una
  Randbedingungen          :: bool list => una
  RandbedingungenBiegung   :: bool list => una
  RandbedingungenNeigung   :: bool list => una
  RandbedingungenMoment    :: bool list => una
  RandbedingungenQuerkraft :: bool list => una
  FunktionsVariable        :: real => una
  Funktionen               :: bool list => una
  Gleichungen              :: bool list => una

  (*Script-names*)
  Biegelinie2Script        :: "[real,real,real,real=>real,bool list,
				bool] => bool"	
	("((Script Biegelinie2Script (_ _ _ _ _ =))// (_))" 9)
  BiegelinieScript         :: "[real,real,real,real=>real,bool list,bool list,
				bool] => bool"	
	("((Script BiegelinieScript (_ _ _ _ _ _ =))// (_))" 9)
  Biege2xIntegrierenScript :: "[real,real,real,bool,real=>real,bool list,
				bool] => bool"		
	("((Script Biege2xIntegrierenScript (_ _ _ _ _ _ =))// (_))" 9)
  Biege4x4SystemScript     :: "[real,real,real,real=>real,bool list,  
				bool] => bool"	
	("((Script Biege4x4SystemScript (_ _ _ _ _ =))// (_))" 9)
  Biege1xIntegrierenScript :: 
	            "[real,real,real,real=>real,bool list,bool list,bool list,
		      bool] => bool"	
	("((Script Biege1xIntegrierenScript (_ _ _ _ _ _ _ =))// (_))" 9)
  Belastung2BiegelScript   :: "[real,real,
	                        bool list] => bool list"	
	("((Script Belastung2BiegelScript (_ _ =))// (_))" 9)
  SetzeRandbedScript       :: "[bool list,bool list,
	                        bool list] => bool list"	
	("((Script SetzeRandbedScript (_ _ =))// (_))" 9)

rules

  Querkraft_Belastung   "Q' x = -q_ x"
  Belastung_Querkraft   "-q_ x = Q' x"

  Moment_Querkraft      "M_b' x = Q x"
  Querkraft_Moment      "Q x = M_b' x"

  Neigung_Moment        "y'' x = -M_b x/ EI"
  Moment_Neigung        "M_b x = -EI * y'' x"

  (*according to rls 'simplify_Integral': .. = 1/a * .. instead .. = ../ a*)
  make_fun_explicit     "Not (a =!= 0) ==> (a * (f x) = b) = (f x = 1/a * b)"

end

