(*. (c) by Richard Lang, 2003 .*)
(* collecting all knowledge for LinearEquations
   created by: rlang 
         date: 02.10
   changed by: rlang
   last change by: rlang
             date: 02.11.04
*)

(* remove_thy"LinEq";
   use_thy"Knowledge/Isac";

   use_thy"Knowledge/LinEq";

   use"ROOT.ML";
   cd"knowledge";
*)

"******* LinEq.ML begin *******";

(*-------------------- theory -------------------------------------------------*)
theory' := overwritel (!theory', [("LinEq.thy",LinEq.thy)]);

(*-------------- rules -------------------------------------------------------*)
val LinEq_prls = (*3.10.02:just the following order due to subterm evaluation*)
  append_rls "LinEq_prls" e_rls 
	     [Calc ("op =",eval_equal "#equal_"),
	      Calc ("Tools.matches",eval_matches ""),
	      Calc ("Tools.lhs"    ,eval_lhs ""),
	      Calc ("Tools.rhs"    ,eval_rhs ""),
	      Calc ("Poly.has'_degree'_in",eval_has_degree_in ""),
 	      Calc ("Poly.is'_polyrat'_in",eval_is_polyrat_in ""),
	      Calc ("Atools.occurs'_in",eval_occurs_in ""),    
	      Calc ("Atools.ident",eval_ident "#ident_"),
	      Thm ("not_true",num_str not_true),
	      Thm ("not_false",num_str not_false),
	      Thm ("and_true",num_str and_true),
	      Thm ("and_false",num_str and_false),
	      Thm ("or_true",num_str or_true),
	      Thm ("or_false",num_str or_false)
              ];
(* ----- erls ----- *)
val LinEq_crls = 
   append_rls "LinEq_crls" poly_crls
   [Thm  ("real_assoc_1",num_str real_assoc_1)
    (*		
     Don't use
     Calc ("HOL.divide", eval_cancel "#divide_"),
     Calc ("Atools.pow" ,eval_binop "#power_"),
     *)
    ];

(* ----- crls ----- *)
val LinEq_erls = 
   append_rls "LinEq_erls" Poly_erls
   [Thm  ("real_assoc_1",num_str real_assoc_1)
    (*		
     Don't use
     Calc ("HOL.divide", eval_cancel "#divide_"),
     Calc ("Atools.pow" ,eval_binop "#power_"),
     *)
    ];

ruleset' := overwritelthy thy (!ruleset',
			[("LinEq_erls",LinEq_erls)(*FIXXXME:del with rls.rls'*)
			 ]);
    
val LinPoly_simplify = prep_rls(
  Rls {id = "LinPoly_simplify", preconds = [], 
       rew_ord = ("termlessI",termlessI), 
       erls = LinEq_erls, 
       srls = Erls, 
       calc = [], 
       (*asm_thm = [],*)
       rules = [
		Thm  ("real_assoc_1",num_str real_assoc_1),
		Calc ("op +",eval_binop "#add_"),
		Calc ("op -",eval_binop "#sub_"),
		Calc ("op *",eval_binop "#mult_"),
		(*  Dont use  
		 Calc ("HOL.divide", eval_cancel "#divide_"),		
		 Calc ("Root.sqrt",eval_sqrt "#sqrt_"),
		 *)
		Calc ("Atools.pow" ,eval_binop "#power_")
		],
       scr = Script ((term_of o the o (parse thy)) "empty_script")
       }:rls);
ruleset' := overwritelthy thy (!ruleset',
			  [("LinPoly_simplify",LinPoly_simplify)]);

(*isolate the bound variable in an linear equation; 'bdv' is a meta-constant*)
val LinEq_simplify = prep_rls(
Rls {id = "LinEq_simplify", preconds = [],
     rew_ord = ("e_rew_ord",e_rew_ord),
     erls = LinEq_erls,
     srls = Erls,
     calc = [],
     (*asm_thm = [("lin_isolate_div","")],*)
     rules = [
	      Thm("lin_isolate_add1",num_str lin_isolate_add1), 
	      (* a+bx=0 -> bx=-a *)
	      Thm("lin_isolate_add2",num_str lin_isolate_add2), 
	      (* a+ x=0 ->  x=-a *)
	      Thm("lin_isolate_div",num_str lin_isolate_div)    
	      (*   bx=c -> x=c/b *)  
	      ],
     scr = Script ((term_of o the o (parse thy)) "empty_script")
     }:rls);
ruleset' := overwritelthy thy (!ruleset',
			[("LinEq_simplify",LinEq_simplify)]);

(*----------------------------- problem types --------------------------------*)
(* 
show_ptyps(); 
(get_pbt ["linear","univariate","equation"]);
*)
(* ---------linear----------- *)
store_pbt
 (prep_pbt LinEq.thy "pbl_equ_univ_lin" [] e_pblID
 (["linear","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["False", (*WN0509 just detected: this pbl can never be used?!?*)
               "Not( (lhs e_) is_polyrat_in v_)",
               "Not( (rhs e_) is_polyrat_in v_)",
               "((lhs e_) has_degree_in v_)=1",
	       "((rhs e_) has_degree_in v_)=1"]),
   ("#Find"  ,["solutions v_i_"]) 
  ],
  LinEq_prls, SOME "solve (e_::bool, v_)",
  [["LinEq","solve_lineq_equation"]]));

(*-------------- methods-------------------------------------------------------*)
store_met
 (prep_met LinEq.thy "met_eqlin" [] e_metID
 (["LinEq"],
   [],
   {rew_ord'="tless_true",rls'=Atools_erls,calc = [], srls = e_rls, prls=e_rls,
    crls=LinEq_crls, nrls=norm_Poly
    (*, asm_rls=[],asm_thm=[]*)}, "empty_script"));

(* ansprechen mit ["LinEq","solve_univar_equation"] *)
store_met
(prep_met LinEq.thy "met_eq_lin" [] e_metID
 (["LinEq","solve_lineq_equation"],
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["Not( (lhs e_) is_polyrat_in v_)",
                "( (lhs e_)  has_degree_in v_)=1"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=LinEq_erls,
    srls=e_rls,
    prls=LinEq_prls,
    calc=[],
    crls=LinEq_crls, nrls=norm_Poly(*,
    asm_rls=[],
    asm_thm=[("lin_isolate_div","")]*)},
    "Script Solve_lineq_equation (e_::bool) (v_::real) =                 \
    \(let e_ =((Try         (Rewrite     all_left            False)) @@  \ 
    \          (Try (Repeat (Rewrite     makex1_x           False))) @@  \ 
    \          (Try         (Rewrite_Set expand_binoms       False)) @@  \ 
    \          (Try (Repeat (Rewrite_Set_Inst [(bdv,v_::real)]           \
    \                                 make_ratpoly_in    False)))    @@  \
    \          (Try (Repeat (Rewrite_Set LinPoly_simplify      False)))) e_;\
    \     e_ = ((Try (Rewrite_Set_Inst [(bdv,v_::real)]                  \
    \                                          LinEq_simplify True)) @@  \
    \            (Repeat(Try (Rewrite_Set LinPoly_simplify     False)))) e_ \
    \ in ((Or_to_List e_)::bool list))"
 ));
"******* LinEq.ML end *******";
get_met ["LinEq","solve_lineq_equation"];
